;;; execution.el ---
;;
;; Copyright (C) 2010 Jan Moringen
;;
;; Author: Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
;; Keywords: task, execution
;;
;; This Program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This Program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program. If not, see <http://www.gnu.org/licenses>.


;;; Commentary:
;;
;; + `execution'
;;   + `execution-blocking'
;;   + `execution-timer'
;;     + `execution-idle'
;;     + `execution-timeout'
;;   + `execution-thread'


;;; History:
;;
;; 0.1 - Initial version


;;; Code:
;;

(require 'tasks/tasks)


;;; Future protocol
;;

(defgeneric future-done-p (future)
  "Return non-nil if the computation represented by FUTURE is done.")

(defgeneric future-get (future &optional timeout)
  "Return the result of FUTURE, waiting if necessary.")

(defgeneric future-cancel (future)
  "Cancel FUTURE.")


;;; Class `execution-result'
;;

(defclass execution-result ()
  ((task        :initarg  :task
		;;:type    task-child
		:documentation
		"")
   (result      :initarg  :result
		:documentation
		"")
   (error       :initarg  :error
		:type     cons
		:reader   result-error
		:documentation
		""))
  "")

(defmethod has-result-p ((this execution-result))
  "Return non-nil if THIS contains a stored a result."
  (slot-boundp this 'result))

(defmethod has-error-p ((this execution-result))
  "Return non-nil if THIS contains a stored error."
  (slot-boundp this 'error))

(defmethod future-done-p ((this execution-result))
  "Return non-nil if the computation represented by THIS is done."
  (or (has-result-p this) (has-error-p this)))


;;; Class `execution'
;;

(defclass execution ()
  ()
  ""
  :abstract t)

(defgeneric execution-execute (execution task)
  "")


;;; Class `execution-blocking-result'
;;

(defclass execution-blocking-result (execution-result)
  ()
  "")

(defmethod initialize-instance ((this execution-blocking-result) slots)
  ""
  (when (next-method-p)
    (call-next-method))

  (with-slots (task result error) this
    (condition-case error-data
	(setq result (execute task))
      (error
       (setq error error-data))))
  )

(defmethod future-done-p ((this execution-blocking-result))
  ""
  t)

(defmethod future-get ((this execution-blocking-result) &optional timeout)
  ""
  (oref this :result))

(defmethod future-cancel ((this execution-blocking-result))
  ""
  (error "canceling not supported"))


;;; Class `execution-blocking'
;;

(defclass execution-blocking (execution)
  ()
  "")

(defmethod execution-execute ((this execution-blocking) task)
  ""
  (execution-blocking-result
   "result"
   :task task))


;;; Class `execution-result-idle'
;;

(defclass execution-result-idle (execution-result)
  ((timer :initarg :timer
	  :type    (or null timer)
	  :documentation
	  ""))
  "")

(defmethod initialize-instance ((this execution-result-idle) slots)
  ""
  (when (next-method-p)
    (call-next-method))

  (with-slots (timer) this
    (setq timer (lexical-let ((this1 this))
		  (run-with-idle-timer
		   1.0 nil
		   (lambda ()
		     (with-slots (task result error timer) this1
		       (condition-case error-data
			   (setq result (execute task))
			 (error
			  (setq error error-data)))
		       (setq timer nil)))))))
  )

(defmethod future-done-p ((this execution-result-idle))
  ""
  (not (oref this :timer)))

(defmethod future-get ((this execution-result-idle) &optional timeout)
  ""
  (while (not (future-done-p this))
    (sit-for 0.1))
  (oref this :result))

(defmethod future-cancel ((this execution-result-idle))
  ""
  (if (future-done-p this)
      (error "too late")
    (with-slots (timer) this
      (cancel-timer timer))))


;;; Class `execution-result-timeout'
;;

;; TODO common superclass for idle, timeout
(defclass execution-result-timeout (execution-result)
  ((timer :initarg  :timer
	  :type     (or null timer)
	  :documentation
	  "")
   (delay :initarg  :delay
	  :type     number
	  :initform 1.0
	  :documentation
	  ""))
  "")

(defmethod initialize-instance ((this execution-result-timeout) slots)
  ""
  (when (next-method-p)
    (call-next-method))

  (with-slots (timer delay) this
    (setq timer (lexical-let ((this1 this))
		  (run-with-timer
		   delay nil
		   (lambda ()
		     (with-slots (task result error timer) this1
		       (condition-case error-data
			   (setq result (execute task))
			 (error
			  (setq error error-data)))
		       (setq timer nil)))))))
  )

(defmethod future-done-p ((this execution-result-timeout))
  ""
  (not (oref this :timer)))

(defmethod future-get ((this execution-result-timeout) &optional timeout)
  ""
  (while (not (future-done-p this))p
    (sit-for 0.1))
  (oref this :result))

(defmethod future-cancel ((this execution-result-timeout))
  ""
  (if (future-done-p this)
      (error "too late")
    (with-slots (timer) this
      (cancel-timer timer))))


;;; Class `execution-timeout'
;;

(defclass execution-timeout (execution)
  ((delay :initarg  :delay
	  :type     number
	  :initform 1.0
	  :documentation
	  ""))
  "")

(defmethod execution-execute ((this execution-timeout) task)
  ""
  (with-slots (delay) this
    (execution-result-timeout
     "result"
     :task  task
     :delay delay)))


;;; Class `execution-result-thread'
;;

(defclass execution-result-thread (execution-result)
  ((thread :initarg :thread
	  :type    (or null thread)
	  :documentation
	  ""))
  "")

(defmethod initialize-instance ((this execution-result-thread) slots)
  ""
  (when (next-method-p)
    (call-next-method))

  (with-slots (thread) this
    (setq thread (lexical-let ((this1 this))
		  (run-with-thread
		   (with-slots (task result error thread) this1
		     (condition-case error-data
			 (setq result (execute task))
		       (error
			(setq error error-data)))
		     (setq thread nil))))))
  )

(defmethod future-done-p ((this execution-result-thread))
  ""
  (not (oref this :thread)))

(defmethod future-get ((this execution-result-thread) &optional timeout)
  ""
  (while (not (future-done-p this))
    (sit-for 0.1))
  (oref this :result))

(defmethod future-cancel ((this execution-result-thread))
  ""
  (if (future-done-p this)
      (error "too late")
    (with-slots (thread) this
      (cancel-thread thread))))


;;; Class `execution-thread'
;;

(defclass execution-thread (execution)
  ()
  "")

(defmethod execution-execute ((this execution-thread) task)
  ""
  (execution-result-thread
   "result"
   :task task))


;;;
;;

(defun tasks-execution-make-execution (class-name &rest args)
  ""
  (let ((class (tasks-execution-select-class class-name)))
    (apply class "execution" args)))

(defvar tasks-execution-classes-alist
  '((blocking . execution-blocking)
    (idle     . execution-idle)
    (timeout  . execution-timeout))
  "")

(when (featurep 'concurrency)
  (add-to-list tasks-execution-classes-alist '(:thread . execution-thread)))

(defun tasks-execution-select-class (name)
  ""
  (cond
   ((null name)
    'execution-blocking)

   ((assoc name tasks-execution-classes-alist)
    (cdr (assoc name tasks-execution-classes-alist)))

   (t
    (signal 'no-such-execution (list name))))
  )

(provide 'tasks/execution)
;;; execution.el ends here
