;;; macros.el ---
;;
;; Copyright (C) 2010 Jan Moringen
;;
;; Author: Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
;; Keywords: tasks, macros, api, syntax, sugar
;;
;; This Program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This Program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program. If not, see <http://www.gnu.org/licenses>.


;;; Commentary:
;;


;;; History:
;;
;; 0.1 - Initial version


;;; Code:
;;

(require 'tasks/tasks)
(require 'tasks/execution)


;;; Macros
;;

(defmacro make-task (name task-classes args
		     &optional doc-string &rest body)
  ""
  (declare (indent     3)
	   (doc-string 3))

  `(tasks-make-task
    ,(symbol-name name) (quote ,task-classes) (quote ,args)
    ,doc-string (quote (progn ,@body))))

(defmacro do-as-task (name task-classes args
		      &optional doc-string &rest body)
  "Run BODY in task NAME and return a future for accessing the result."
  (declare (indent     3)
	   (doc-string 3)
	   (debug      (symbolp list list
				[ &optional stringp ] def-body)))

  `(future-get (run-as-task ,name ,task-classes ,args
		 ,doc-string ,@body)))

(defmacro run-as-task (name task-classes args
		       &optional doc-string &rest body) ;; TODO can we construct the task at expansion time?
  "Run BODY in task NAME waiting for the result and returning it."
  (declare (indent     3)
	   (doc-string 3)
	   (debug      (symbolp list list
				[ &optional stringp ] def-body)))

  (let ((task-var      (make-symbol "task"))
	(execution-var (make-symbol "execution"))
	(task-args     (tasks-macros-extract-task-args args)))
    (destructuring-bind (execution-class execution-args)
	(tasks-macros-parse-execution (plist-get args :execution))
      `(let ((,task-var      (tasks-make-task
			      ,(symbol-name name)
			      (quote ,task-classes)
			      (quote ,task-args)
			      ,doc-string (quote (progn ,@body))))

	     (,execution-var (apply #'tasks-execution-make-execution
				    (quote ,execution-class)
				    (list ,@execution-args))))
	 (execution-execute ,execution-var ,task-var))))
  )


;;; Utility Functions
;;

(defun tasks-macros-extract-task-args (args)
  ""
  (let ((result)
	(rest args))
    (while rest
      (unless (eq (first rest) :execution)
	(push (nth 0 rest) result)
	(push (nth 1 rest) result))
      (setq rest (cddr rest)))
    (reverse result))
  )

(defun tasks-macros-parse-execution (spec)
  ""
  (cond
   ((symbolp spec)
    (list spec nil))

   ((list spec)
    (list (first spec) (rest spec)))

   (t
    (error "invalid execution spec")))
  )

(provide 'tasks/macros)


;;; Unit Tests
;;

(eval-when-compile
  (when (require 'ert nil t)

    (ert-deftest tasks/macros-test-make-task-macroexpand ()
      ""
      ;; Minimal: no task or execution specification
      (should (equal
	       (macroexpand
		'(make-task add-numbers () ()
		   "Add some numbers"
		   (+ 1 2 3)))

	       '(tasks-make-task
		 "add-numbers"
		 '()
		 '()
		 "Add some numbers"
		 (quote (progn (+ 1 2 3))))))

      ;; With task specification
      (should (equal
	       (macroexpand
		'(make-task add-numbers (progress) ()
		   "Add some numbers"
		   (+ 1 2 3)))

	       '(tasks-make-task
		 "add-numbers"
		 '(progress)
		 '()
		 "Add some numbers"
		 (quote (progn (+ 1 2 3))))))
      )

    (ert-deftest tasks/macros-test-do-as-task-macroexpand ()
      (should
       (equal
	(macroexpand
	 '(do-as-task add-numbers
	      (progress)
	      (:bli 5 :bla t :execution (timeout :delay 5))
	    "Add some numbers"
	    (+ 1 2 3)))

	'(future-get
	  (run-as-task add-numbers
	      (progress)
	      (:bli 5 :bla t :execution (timeout :delay 5))
	    "Add some numbers"
	    (+ 1 2 3)))))
      )

    (ert-deftest tasks/macros-test-do-as-task-smoke ()
      ""
      (do-as-task add-numbers (progress) (:execution blocking)
	""
	(dolist (i (number-sequence 1 1000))
	  (+ 1 2 3)
	  (progress (/ (float i) (float 1000))))
	123)
      )

    (ert-deftest tasks/macros-test-run-as-task-macroexpand ()
      ""
      (should
       (equal
	(macroexpand
	 '(run-as-task add-numbers
	      (progress)
	      (:bli 5 :bla t :execution (timeout :delay 5))
	    "Add some numbers"
	    (+ 1 2 3)))

	'(let ((task      (tasks-make-task
			   "add-numbers"
			   '(progress)
			   '(:bli 5 :bla t)
			   "Add some numbers"
			   '(progn (+ 1 2 3))))
	       (execution (apply
			   #'tasks-execution-make-execution
			   'timeout (list :delay 5))))
	   (execution-execute execution task))))
      )

    (ert-deftest tasks/macros-test-run-as-task-smoke ()
      ""
      (run-as-task idle-numbers (cancelable progress) (:execution timeout)
	"Compute some numbers during idle time."
	(dolist (i (number-sequence 1 1000))
	  (+ 1 2 3)
	  (progress (/ (float i) (float 1000)))
	  (maybe-cancel 1))
	1)
      )

    ))

;;; macros.el ends here
