;;; tasks.el --- Task classes
;;
;; Copyright (C) 2010 Jan Moringen
;;
;; Author: Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
;; Keywords: tasks, concurrency
;;
;; This Program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This Program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program. If not, see <http://www.gnu.org/licenses>.


;;; Commentary:
;;
;; What a task is
;; + What gets executed
;; + Execution context (functions callable from task code)
;; + Meta information (name, description)
;; What a task is not
;; + how to execute (synchronous vs. asynchronous)
;; + when to execute (delayed, lazy etc)
;;
;; + `task'
;;   + `progress-task'
;;   + `cancelable-task'
;;     + `cooperative-cancelable-task'


;;; History:
;;
;; 0.1 - Initial version


;;; Code:
;;

(require 'eieio)

(require 'tasks/hook-object)


;;; Class `task'
;;

(defclass task (hook-object) ;; TODO do we want all tasks to be hook objects?
  ((code                :initarg  :code
			:documentation
			"The code executed by this task.")
   (lambda              :initarg  :lambda
			:documentation
			"A lambda expression that wraps the
original code of the tasks in an environment that provides
additional functionalities. The particular things contained in
this environment depend on the `compute-task-lambda' method of
the task class.")
   (description         :initarg  :description
			:type     (or null string)
			:reader   task-description
			:documentation
			"A textual description of the task.")
   ;; Hooks
   (before-execute-hook :initarg  :before-execute-hook
			:type     list
			:initform nil
			:documentation
			"This hook is run when execution of the
task starts.")
   (after-execute-hook  :initarg  :after-execute-hook
			:type     list
			:initform nil
			:documentation
			"This hook is run when execution of the
task finishes. TODO what about canceled tasks?"))
  "")

(defmethod execute ((this task))
  ""
  (with-slots (code lambda) this
    ;; Compute and store lambda if necessary.
    (when (not (slot-boundp this :lambda))
      (oset this :lambda `(lambda () ,(compute-task-lambda this code))))

    ;; Execute stored lambda and run hooks.
    (prog2
	;; Run before execute hook
	(object-run-hook-with-args this 'before-execute-hook)

	;; Run the actual task code
	(funcall lambda)

      ;; Run after execute hook
      (object-run-hook-with-args this 'after-execute-hook)))
  )

(defmethod compute-task-lambda ((this task) code)
  ""
  code)

(defmethod short-description ((this task) &optional max-length)
  ""
  (with-slots (description) this
    (substring description 0 (min
			      (or
			       (position ?\n description)
			       max-length
			       (length description))
			      (length description)))))

(defmethod object-print ((this task) &rest strings)
  "Return a textual representation of THIS."
  (when (next-method-p) ;; TODO
    (apply #'call-next-method this
	   (format " \"%s\"" (short-description this))
	   strings)))


;;; Class `progress-task'
;;

(defclass progress-task (task)
  ((progress      :initarg  :progress
		  :type     (or boolean (number 0 1))
		  :reader   task-progress
		  :initform nil
		  :documentation
		  "")
   ;; Hooks
   (progress-hook :initarg  :progress-hook
		  :type     list
		  :initform nil
		  :documentation
		  "This hook is run when the execution of the
task progressed. Functions which are added to this hook have to
take two arguments: the task object and the progress."))
  "")

(defmethod compute-task-lambda ((this progress-task) code)
  ""
  `(flet ((progress (current)
	    (handle-progress ,this current)))
     ,(if (next-method-p)
	  (call-next-method this code)
	`(progn ,,@code))))

(defmethod handle-progress ((this progress-task) progress)
  ""
  (oset this :progress progress)
  (object-run-hook-with-args this 'progress-hook progress))

(defmethod object-print ((this progress-task) &rest strings)
  ""
  (when (next-method-p)
    (with-slots (progress) this
      (apply #'call-next-method
	     this
	     (case progress
	       ('t
		" completed")

	       ('nil
		" progress ?")

	       (t
		(format " progress %.0f %%"
			(* 100.0 (float progress)))))
	     strings)))
  )


;;; Class `cancelable-task'
;;

(defclass cancelable-task (task)
  ()
  ""
  :abstract t)

(defmethod cancel ((this cancelable-task))
  ""
  )


;;; Class `cooperative-cancelable-task'
;;

(defclass cooperative-cancelable-task (cancelable-task)
  ((cancel-symbol :initarg :cancel-symbol
		  :type    symbol
		  :documentation
		  "")
   (canceled      :initarg :canceled
		  :type    boolean
		  :documentation
		  ""))
  "")

(defmethod initialize-instance ((this cooperative-cancelable-task) slots)
  ""
  (when (next-method-p)
    (call-next-method))

  (oset this :cancel-symbol (make-symbol "cancel-symbol"))
  )

(defmethod compute-task-lambda ((this cooperative-cancelable-task) code)
  ""
  (with-slots (cancel-symbol) this
    `(flet ((maybe-cancel (&optional data)
	      (when (oref ,this :canceled)
		(throw (quote ,cancel-symbol) data))))
       ,(if (next-method-p)
	    (call-next-method this code)
	  `(progn ,,@code))))
  )

(defmethod execute ((this cooperative-cancelable-task))
  ""
  (with-slots (cancel-symbol canceled) this
    (setq canceled nil)
    (catch cancel-symbol
      (when (next-method-p)
	(call-next-method)))))

(defmethod cancel ((this cooperative-cancelable-task))
  ""
  (oset this :canceled t))


;;; Construction
;;

(defvar tasks-task-classes-alist
  '((progress   . progress-task)
    (cancelable . cooperative-cancelable-task))
  "")

(defvar tasks-task-classes (make-hash-table :test 'equal)
  "")

(defun tasks-make-task (name class-names args description code)
  ""
  (let* ((classes (mapcar #'tasks-select-task-class class-names))
	 (class   (tasks-ensure-task-class classes)))
    (funcall
     class
     name
     :code        code
     :description description))
  )

(defun tasks-select-task-class (name)
  ""
  (cdr (assoc name tasks-task-classes-alist)))

(defun tasks-ensure-task-class (superclasses)
  ""
  (let ((sorted (sort* superclasses #'string< :key #'symbol-name)))
    (or (gethash sorted tasks-task-classes)
	(puthash sorted (tasks-make-task-class superclasses) tasks-task-classes)))
  )

(defun tasks-make-task-class (superclasses)
  ""
  (let ((class-name (intern (symbol-name (gensym)))))
    (eval
     `(defclass ,class-name ,superclasses
	()
	"TODO"
	:method-invocation-order :c3
	))
    class-name)
  )

(provide 'tasks/tasks)


;;; Unit Tests
;;

(eval-when-compile
  (when (require 'ert nil t)

    ;; Tests for class `task'

    (ert-deftest tasks/tasks-test-task-smoke ()
      "Smoke test for class `task'."
      (let ((task (task "bla" :code '(+ 1 2 3))))
	(should (= (execute task) 6)))
      )

    (ert-deftest tasks/tasks-test-task-compute-task-lambda ()
      "Test `compute-task-lambda' for the `task' class."
      (let ((task (task "test" :code '(+ 1 2))))
	(should (equal
		 (compute-task-lambda task (oref task :code))
		 '(+ 1 2))))

      (let ((task (task "test" :code '(progn (+ 1 2) (+ 3 4)))))
	(should (equal
		 (compute-task-lambda task (oref task :code))
		 '(progn (+ 1 2) (+ 3 4)))))
      )

    ;; Tests for class `progress-task'

    (ert-deftest tasks/tasks-test-progress-task-smoke ()
      "Smoke test for class `progress-task'."
      nil)

    (ert-deftest tasks/tasks-test-progress-task-object-print ()
      "Smoke test for class `progress-task'."
      (let ((task (progress-task
		   "bla"
		   :code '(+ 1 2 3)
		   :description "blup")))
	;; Initial
	(should (string= (object-print task)
			 "#<progress-task bla \"blup\" progress ?>"))

	;; Completed
	(oset task :progress t)
	(should (string= (object-print task)
			 "#<progress-task bla \"blup\" completed>"))

	;; Numeric
	(oset task :progress 0.8)
	(should (string= (object-print task)
			 "#<progress-task bla \"blup\" progress 80 %>")))
      )

    ;; Tests for class `cooperative-cancelable-task'

    (ert-deftest tasks/tasks-test-cooperative-cancelable-task-smoke ()
      "Smoke test for class `cooperative-cancelable-task'."
      (let ((task (cooperative-cancelable-task
		   "test"
		   :code '(progn
			    (maybe-cancel (quote canceled))
			    (quote finished)))))
	(should (eq (execute task) 'finished)))
      )

    (ert-deftest tasks/tasks-test-cooperative-cancelable-task-compute-task-lambda ()
      "Test the `compute-task-lambda' method for the `cooperative-cancelable-task' class."
      (let ((task (cooperative-cancelable-task
		   "test"
		   :code '(maybe-cancel 1))))
	(should
	 (equal
	  (compute-task-lambda task (oref task :code))
	  `(flet ((maybe-cancel (&optional data)
		    (when (oref ,task :canceled)
		      (throw 'cancel-symbol data))))
	     (maybe-cancel 1)))))
      )

    ))

;;; tasks.el ends here
